/*
 * Embeddable Scheme Interpreter v0.9
 * Copyright (c) 2010-2012 by Andrej Bukosek.  All rights reserved.
 *
 * Released under the 3-clause BSD license (see file LICENSE).
 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>

#include <math.h>
#include <float.h>
#include <gc.h>

#include "scheme.h"


/**
 *
 * Platform compatibility
 *
 **/

#ifndef GC_STRDUP
#define GC_STRDUP(s) GC_strdup(s)

/*
 * GC_strdup() -- For libraries which are too old to have it.
 */
static char *GC_strdup(const char *s)
{
	size_t  len = strlen(s) + 1;
	char   *ret = (char *)GC_MALLOC(len);

	strcpy(ret, s);

	return ret;
}
#endif /* !GC_STRDUP */

#ifdef __linux
/*
 * Include implementations of strlcpy() and strlcat() from OpenBSD.
 *
 * Copyright (c) 1998 Todd C. Miller <Todd.Miller@courtesan.com>
 */

/*
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
size_t strlcpy(char *dst, const char *src, size_t siz)
{
	char *d = dst;
	const char *s = src;
	size_t n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) {
		do {
			if ((*d++ = *s++) == 0)
				break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (siz != 0)
			*d = '\0';              /* NUL-terminate dst */
		while (*s++)
			;
	}

	return(s - src - 1);    /* count does not include NUL */
}

/*
 * Appends src to string dst of size siz (unlike strncat, siz is the
 * full size of dst, not space left).  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz <= strlen(dst)).
 * Returns strlen(src) + MIN(siz, strlen(initial dst)).
 * If retval >= siz, truncation occurred.
 */
size_t strlcat(char *dst, const char *src, size_t siz)
{
	char *d = dst;
	const char *s = src;
	size_t n = siz;
	size_t dlen;

	/* Find the end of dst and adjust bytes left but don't go past end */
	while (n-- != 0 && *d != '\0')
		d++;
	dlen = d - dst;
	n = siz - dlen;

	if (n == 0)
		return(dlen + strlen(s));
	while (*s != '\0') {
		if (n != 1) {
			*d++ = *s;
			n--;
		}
		s++;
	}
	*d = '\0';

	return(dlen + (s - src));       /* count does not include NUL */
}
#endif



/**
 *
 * Error handling
 *
 **/

/*
 * err() -- Assemble the error string and jump to the error handler.
 */
void err(context *ctx, enum err_type type, const char *fmt, ...)
{
	char    buf[512];
	va_list ap;

	va_start(ap, fmt);
	switch (type) {
		case ERR_FATAL:
			vsnprintf(buf, sizeof(buf), fmt, ap);
			snprintf(ctx->error, sizeof(ctx->error), "FATAL ERROR: %s\n", buf);

			va_end(ap);
			longjmp(ctx->err_handler, ERR_FATAL);
			break;


		case ERR_SYNTAX:
			vsnprintf(ctx->p->error, sizeof(ctx->p->error), fmt, ap);

			/* If possible, let the user know where the error occurred */
			if (ctx->p->error[0] != '\0') {
				const char *c = ctx->p->line_start;
				size_t      i, extraspace = 0;

				snprintf(buf, sizeof(buf), "SYNTAX ERROR: Line %lu: %s:\n",
						 ctx->p->line, ctx->p->error);

				/* Print offending line */
				while (*c != '\n' && c < ctx->p->source + ctx->p->length) {
					/* Handle tabs */
					if (*c == '\t') {
						extraspace += 3;
						strlcat(buf, "    ", sizeof(buf));
					} else {
						char ch[2];

						ch[0] = *c;
						ch[1] = '\0';

						strlcat(buf, ch, sizeof(buf));
					}

					c++;
				}
				strlcat(buf, "\n", sizeof(buf));

				/* Point to the character at which the error occurred */
				for (i = 0; i < ctx->p->column + extraspace; i++)
					strlcat(buf, " ", sizeof(buf));

				strlcat(buf, "^\n", sizeof(buf));
			}

			strlcpy(ctx->error, buf, sizeof(ctx->error));

			va_end(ap);
			longjmp(ctx->err_handler, ERR_SYNTAX);
			break;


		case ERR_RUNTIME:
			vsnprintf(buf, sizeof(buf), fmt, ap);

			/* If possible, let the user know where the error occurred */
			if (ctx->p && !(ctx->err_line == 0 && ctx->err_column == 0)) {
				const char *c = ctx->p->source;
				size_t      i = 0, extraspace = 0;

				snprintf(ctx->error, sizeof(ctx->error),
						 "RUNTIME ERROR: Line %lu: %s:\n",
						 ctx->err_line, buf);

				/* Seek to beginning of specified line */
				while (c < ctx->p->source + ctx->p->length) {
					if (i == ctx->err_line - 1)
						break;

					if (*c == '\n')
						i++;

					c++;
				}

				/* Print offending line */
				while (*c != '\n' && c < ctx->p->source + ctx->p->length) {
					/* Handle tabs */
					if (*c == '\t') {
						extraspace += 3;
						strlcat(ctx->error, "    ", sizeof(ctx->error));
					} else {
						char ch[2];

						ch[0] = *c;
						ch[1] = '\0';

						strlcat(ctx->error, ch, sizeof(ctx->error));
					}

					c++;
				}
				strlcat(ctx->error, "\n", sizeof(ctx->error));

				/* Point to the character at which the error occurred */
				for (i = 0; i < ctx->err_column + extraspace; i++)
					strlcat(ctx->error, " ", sizeof(ctx->error));

				strlcat(ctx->error, "^\n", sizeof(ctx->error));
			} else {
				snprintf(ctx->error, sizeof(ctx->error),
						 "RUNTIME ERROR: %s\n", buf);
			}

			va_end(ap);
			longjmp(ctx->err_handler, ERR_RUNTIME);
			break;
	}
	va_end(ap);
}



/**
 *
 * Object allocation & constructors
 *
 **/

/*
 * create_obj() -- Allocate memory for an object of the specified type.
 */
static obj *create_obj(context *ctx, enum obj_type type)
{
	const size_t size = (type == NUM)       ? sizeof(obj_num)  :
						(type == SYM)       ? sizeof(obj_sym)  :
						(type == STR)       ? sizeof(obj_str)  :
						(type == CONS)      ? sizeof(obj_cons) :
						(type == LAMBDA)    ? sizeof(obj_proc) :
						(type == PRIMITIVE) ? sizeof(obj_prim) :
						sizeof(obj);

	/* Atomic objects don't contain pointers to other objects */
	const int atomic = (type == NIL || type == T || type == F || type == NUM);

	/* Allocate memory for the object */
	obj *o = atomic ? (obj *) GC_MALLOC_ATOMIC(size) : (obj *) GC_MALLOC(size);

	if (!o)
		err(ctx, ERR_FATAL, "Out of memory");

	memset(o, 0, size);
	o->type = type;

	/* Save source code position for nicer runtime error printing */
	if (ctx->p && ctx->p->ptr < ctx->p->source + ctx->p->length) {
		o->at_line   = ctx->p->line;
		o->at_column = ctx->p->column;
	}

	return o;
}


obj *cons(context *ctx, obj *_car, obj *_cdr)
{
	obj *o = create_obj(ctx, CONS);
	setcar(o, _car);
	setcdr(o, _cdr);
	return o;
}

obj *mknil(context *ctx)
{
	return create_obj(ctx, NIL);
}

obj *mkt(context *ctx)
{
	return create_obj(ctx, T);
}

obj *mkf(context *ctx)
{
	return create_obj(ctx, F);
}

obj *mknum(context *ctx, double n)
{
	obj *o = create_obj(ctx, NUM);
	numval(o) = n;
	return o;
}

obj *mksym(context *ctx, const char *sym)
{
	obj *o = create_obj(ctx, SYM);
	symname(o) = GC_STRDUP(sym);
	return o;
}

obj *mkstr(context *ctx, const char *str)
{
	obj *o = create_obj(ctx, STR);
	strval(o) = GC_STRDUP(str);
	return o;
}

obj *mkprimitive(context *ctx, primitive prim)
{
	obj *o = create_obj(ctx, PRIMITIVE);
	primval(o) = prim;
	return o;
}

obj *mkproc(context *ctx, obj *args, obj *code, obj *env)
{
	obj *o = create_obj(ctx, LAMBDA);
	procargs(o) = args;
	proccode(o) = code;
	procenv(o)  = env;
	return o;
}



/**
 *
 * Symbol handling
 *
 **/

/*
 * findsym() -- Locate specified symbol in the context's symbol list.
 */
static obj *findsym(context *ctx, const char *name)
{
	obj *list = ctx->symbols;

	while (!isnil(list)) {
		if (!strcmp(name, symname(car(list))))
			return car(list);

		list = cdr(list);
	}

	return ctx->nil;
}


/*
 * intern() -- Add symbol to context's symbol list if it's not there yet.
 */
static obj *intern(context *ctx, const char *name)
{
	/* First check if the symbol already exists */
	obj *sym = findsym(ctx, name);

	/* Symbol already exists, return it */
	if (!isnil(sym))
		return sym;

	/* Symbol doesn't exist yet, create and add it to the symbol list */
	sym = mksym(ctx, name);
	ctx->symbols = cons(ctx, sym, ctx->symbols);

	return sym;
}



/**
 *
 * Environment handling
 *
 **/

/*
 * extend() -- Add a new (sym val) pair to the beginning of env.
 */
#define extend(CTX, ENV, SYM, VAL) \
	(cons((CTX), cons((CTX), (SYM), (VAL)), (ENV)))

/*
 * multiple_extend() -- Add multiple new (sym val) pairs to specified env.
 */
static obj *multiple_extend(context *ctx, obj *env, obj *syms, obj *vals)
{
	return isnil(syms) ?
				env :
				multiple_extend(ctx, extend(ctx, env, car(syms), car(vals)),
								cdr(syms), cdr(vals));
}

/*
 * extend_top() -- Add a new (sym val) pair to top (global) environment.
 */
static obj *extend_top(context *ctx, obj *sym, obj *val)
{
	setcdr(ctx->top_env, cons(ctx, cons(ctx, sym, val), cdr(ctx->top_env)));
	return val;
}

/*
 * assoc() -- Locate value of specified key in the specified associative list.
 */
obj *assoc(context *ctx, obj *key, obj *alist)
{
	if (isnil(alist))
		return ctx->nil;

	if (car(car(alist)) == key)
		return car(alist);

	return assoc(ctx, key, cdr(alist));
}



/**
 *
 * Lexer
 *
 **/

/*
 * next_token() -- Return next token in parser stream.
 */
static enum tok_type next_token(context *ctx)
{
	parser *p  = ctx->p;
	char    ch = *(p->ptr);

	/* Save current position */
	p->ptr_prev        = p->ptr;
	p->line_start_prev = p->line_start;
	p->line_prev       = p->line;
	p->column_prev     = p->column;

again:

	/* Handle whitespace */
	while (isspace(ch) && (p->ptr < p->source + p->length)) {
		if (ch == '\n') {
			p->line++;
			p->column = 0;
			p->line_start = p->ptr + 1;
		} else {
			p->column++;
		}

		p->ptr++;
		ch = *(p->ptr);
	}

	/* Handle comments */
	if (ch == ';') {
		while (ch != '\n' && (p->ptr < p->source + p->length)) {
			p->ptr++;
			ch = *(p->ptr);
		}

		p->line++;
		p->column = 0;
		p->line_start = p->ptr + 1;

		p->ptr++;
		ch = *(p->ptr);

		goto again;
	}

	if (p->ptr >= p->source + p->length)
		return TOK_EOI;

	/* Handle single-char tokens and constants */
	switch (ch) {
		case '(':  p->ptr++; p->column++; return TOK_LPAR;
		case ')':  p->ptr++; p->column++; return TOK_RPAR;
		case '\'': p->ptr++; p->column++; return TOK_QUOTE;
		case '.':  p->ptr++; p->column++; return TOK_DOT;
		case '#':
			p->ptr++; p->column++;

			if (*(p->ptr) == 't') {
				p->ptr++; p->column++;
				return TOK_T;
			} else if (*(p->ptr) == 'f') {
				p->ptr++; p->column++;
				return TOK_F;
			}

			err(ctx, ERR_SYNTAX, "Invalid constant, expected 't' or 'f'");
			break;
	}

	if (isdigit(ch) || ((ch == '-' || ch == '+') && isdigit(*(p->ptr + 1)))) {
		/* Parse a number */
		char *end = NULL;

		p->tok_num = strtod(p->ptr, &end);

		p->column += end - p->ptr;
		p->ptr     = end;

		if (!isspace(*p->ptr) && (*p->ptr) != ')' && (*p->ptr) != 0)
			err(ctx, ERR_SYNTAX, "Malformed number");

		return TOK_NUM;
	}

	if (isalpha(ch) || strchr("-?!+*/=<>:@&%^_~", ch) != NULL) {
		/* Parse a symbol */
		const char *end = p->ptr;
		size_t      len;

		while ((isalnum(*end) || strchr("-?!+*/=<>:@&%^_~", *end) != NULL) &&
			   (end < p->source + p->length)) {
			end++;
			p->column++;
		}

		len = (size_t)(end - p->ptr + 1);

		if (len >= sizeof(p->tok_sym)) {
			const size_t over = len - sizeof(p->tok_sym) + 1;

			err(ctx, ERR_SYNTAX, "Symbol is %d character%s too long",
				over, over > 1 ? "s" : "");
		}

		memcpy(p->tok_sym, p->ptr, len - 1);
		p->tok_sym[len - 1] = '\0';

		p->ptr = end;

		return TOK_SYM;
	}

	if (ch == '\"') {
		/* Parse string */
		const char *end = p->ptr + 1;
		size_t      len;

		while ((end < p->source + p->length) && *end != '\"') {
			if (*end == '\\' && *(end + 1) == '\"') {
				end++;
				p->column++;
			}

			end++;
			p->column++;
		}

		if (end >= p->source + p->length)
			err(ctx, ERR_SYNTAX, "Unterminated string");

		len = (size_t)(end - p->ptr);

		p->tok_str = (char *)GC_MALLOC(len + 1);
		if (!p->tok_str)
			err(ctx, ERR_FATAL, "Cannot allocate memory for string");

		memcpy(p->tok_str, p->ptr + 1, len - 1);
		p->tok_str[len - 1] = '\0';

		p->ptr = end + 1;

		return TOK_STR;
	}

	/* Parse error */
	err(ctx, ERR_SYNTAX, "Unexpected character");
	return TOK_EOI;
}


static void putback_token(context *ctx)
{
	parser *p = ctx->p;

	/* Restore previously saved position */
	p->ptr        = p->ptr_prev;
	p->line_start = p->line_start_prev;
	p->line       = p->line_prev;
	p->column     = p->column_prev;
}



/**
 *
 * Parser
 *
 **/

static obj *parse_list(context *ctx);

static obj *parse_obj(context *ctx)
{
	enum tok_type t = next_token(ctx);

	if (t == TOK_EOI)
		return ctx->nil;

	switch (t) {
		case TOK_LPAR:  return parse_list(ctx);
		case TOK_NUM:   return mknum(ctx, ctx->p->tok_num);
		case TOK_SYM:   return intern(ctx, ctx->p->tok_sym);
		case TOK_STR:   return mkstr(ctx, ctx->p->tok_str);
		case TOK_T:     return mkt(ctx);
		case TOK_F:     return mkf(ctx);
		case TOK_QUOTE: return cons(ctx, ctx->s_quote,
									cons(ctx, parse_obj(ctx), ctx->nil));
	}

	putback_token(ctx);
	err(ctx, ERR_SYNTAX, "Unexpected token");
	return ctx->nil;
}


static obj *parse_list(context *ctx)
{
	enum tok_type  t = next_token(ctx);
	obj           *tmp, *ret;

	if (t == TOK_RPAR)
		return ctx->nil;

	if (t == TOK_DOT) {
		tmp = parse_obj(ctx);

		if (next_token(ctx) != TOK_RPAR) {
			putback_token(ctx);
			err(ctx, ERR_SYNTAX, "Missing ')'");
			return ctx->nil;
		}

		return tmp;
	}

	if (t == TOK_EOI) {
		putback_token(ctx);
		err(ctx, ERR_SYNTAX, "Unexpected end of input");
		return ctx->nil;
	}

	putback_token(ctx);

	tmp = parse_obj(ctx);
	ret = cons(ctx, tmp, parse_list(ctx));

	putback_token(ctx);

	if (next_token(ctx) != TOK_RPAR) {
		putback_token(ctx);
		err(ctx, ERR_SYNTAX, "Missing ')'");
		return ctx->nil;
	}

	return ret;
}


/*
 * parse() -- Parse source code and generate a list for eval().
 */
obj *parse(context *ctx, const char *src)
{
	parser *p;
	obj    *ret, *node, *cur;

	if (ctx->p) {
		GC_FREE(ctx->p);
		ctx->p = NULL;
	}

	p = (parser *)GC_MALLOC(sizeof(parser));
	if (!p)
		err(ctx, ERR_FATAL, "Out of memory");

	p->source     = src;
	p->ptr        = p->source;
	p->ptr_prev   = NULL;
	p->line_start = p->source;
	p->error[0]   = '\0';
	p->line       = 1;
	p->column     = 0;
	p->length     = strlen(p->source);

	ctx->p = p;

	ret = cons(ctx, ctx->s_begin, cons(ctx, parse_obj(ctx), ctx->nil));

	cur = cdr(ret);

	while (!isnil(node = parse_obj(ctx))) {
		setcdr(cur, cons(ctx, node, ctx->nil));
		cur = cdr(cur);
	}

	p->line = p->column = 0;

	/*GC_FREE(p);
	ctx->p = NULL;*/

	return ret;
}


void display(context *ctx, FILE *ofp, obj *op)
{
	if (!op)
		return;

	switch (op->type) {
		case NIL:       fprintf(ofp, "()");                break;
		case T:         fprintf(ofp, "#t");                break;
		case F:         fprintf(ofp, "#f");                break;
		case NUM:       fprintf(ofp, "%.16g", numval(op)); break;
		case SYM:       fprintf(ofp, "%s", symname(op));   break;
		case STR:       fprintf(ofp, "%s", strval(op));    break;
		case LAMBDA:    fprintf(ofp, "#<lambda>");         break;
		case PRIMITIVE: fprintf(ofp, "#<primitive>");      break;
		case CONS:
			fprintf(ofp, "(");
			for (;;) {
				display(ctx, ofp, car(op));
				if (isnil(cdr(op))) {
					fprintf(ofp, ")");
					break;
				}
				op = cdr(op);
				if (op->type != CONS) {
					fprintf(ofp, " . ");
					display(ctx, ofp, op);
					fprintf(ofp, ")");
					break;
				}
				fprintf(ofp, " ");
			}
			break;
	}
}



/**
 *
 * Argument checking
 *
 **/

void check_arg_count(context *ctx, const char *fn, obj *args, int min, int max)
{
	obj *o = args;
	int  n = 0;

	if (min == -1 && max == -1)
		return;

	while (!isnil(o) && iscons(o)) {
		n++;
		o = cdr(o);

		if (max != -1 && n > max)
			break;
	}

	if (min != -1 && n < min)
		err(ctx, ERR_RUNTIME,
			"Too few arguments for '%s' given, expected %s %d",
			fn, (min == max) ? "exactly" : "at least", min);

	if (max != -1 && n > max)
		err(ctx, ERR_RUNTIME,
			"Too many arguments for '%s' given, expected %s %d",
			fn, (min == max) ? "exactly" : "at most", max);
}


void check_arg_types(context *ctx, const char *fn, obj *args, ...)
{
	va_list  ap;
	int      n = 0;
	obj     *o = args;

	va_start(ap, args);

	while (!isnil(o) && iscons(o)) {
#ifdef __sgi
/* Turn off "conversion from pointer to same-sized integral type" for va_arg */
#pragma set woff 3970
#endif
		enum obj_type ot = car(o)->type;					/* obj type */
		enum obj_type et = (enum obj_type) va_arg(ap, int);	/* expected type */
#ifdef __sgi
#pragma reset woff 3970
#endif

		n++;

		if (ot != et) {
			const char *typestr, *spec;

			/* Generate type string */
			if      (et == NIL)       typestr = "nil";
			else if (et == T)         typestr = "boolean";
			else if (et == F)         typestr = "boolean";
			else if (et == NUM)       typestr = "number";
			else if (et == SYM)       typestr = "symbol";
			else if (et == STR)       typestr = "string";
			else if (et == CONS)      typestr = "list";
			else if (et == LAMBDA)    typestr = "lambda";
			else if (et == PRIMITIVE) typestr = "primitive";
			else                      typestr = "<UNKNOWN>";

			/* Generate ordinal number specifier */
			if      ((n % 10) == 1 && ((n - 1) % 100) != 10) spec = "st";
			else if ((n % 10) == 2 && ((n - 2) % 100) != 10) spec = "nd";
			else if ((n % 10) == 3 && ((n - 3) % 100) != 10) spec = "rd";
			else                                             spec = "th";

			err(ctx, ERR_RUNTIME,
				"Incorrect type for %d%s argument of '%s', %s expected",
				n, spec, fn, typestr);
		}

		o = cdr(o);
	}

	va_end(ap);
}


/*
 * list_length() -- Return length of specified list.
 */
int list_length(obj *list)
{
	int n = 0;

	while (!isnil(list)) {
		n++;
		list = cdr(list);
	}

	return n;
}



/**
 *
 * Evaluation
 *
 **/

static obj *eval_list(context *ctx, obj *exps, obj *env);

obj *eval(context *ctx, obj *exp, obj *env)
{
	obj *tmp = NULL, *proc = NULL, *vals = NULL;

	ctx->error[0] = 0;

eval_start:
	if (isnil(exp))
		return ctx->nil;

	ctx->err_line   = exp->at_line;
	ctx->err_column = exp->at_column;

	switch (exp->type) {
		/* Handle straightforward types first */
		case NUM:
		case STR:
		case LAMBDA:
		case PRIMITIVE:
		case T:
		case F:
			return exp;

		/* Symbols need a lookup */
		case SYM:
			tmp = assoc(ctx, exp, env);			/* Get sym->val pair */

			if (isnil(tmp))
				err(ctx, ERR_RUNTIME, "Unbound symbol '%s'", symname(exp));

			return cdr(tmp);					/* Return symbol value */

		/* For everything else, there's cons */
		case CONS:
			/* (quote {exp}) */
			if (car(exp) == ctx->s_quote) {
				check_arg_count(ctx, "quote", cdr(exp), 1, 1);

				return car(cdr(exp));
			}

			/* (define {sym} {exp}) */
			if (car(exp) == ctx->s_define) {
				check_arg_count(ctx, "define", cdr(exp), 2, 2);

				if (!issym(car(cdr(exp))))
					err(ctx, ERR_RUNTIME,
						"Incorrect type for 1st argument of 'define', symbol expected");

				return extend_top(ctx, car(cdr(exp)),
								  eval(ctx, car(cdr(cdr(exp))), env));
			}

			/* (lambda ([args]) {exp}) */
			if (car(exp) == ctx->s_lambda) {
				check_arg_count(ctx, "lambda", cdr(exp), 2, 2);

				return mkproc(ctx, car(cdr(exp)), cdr(cdr(exp)), env);
			}

			/* (if {bool_exp} {true_exp} {false_exp}) */
			if (car(exp) == ctx->s_if) {
				obj *ifexp;

				check_arg_count(ctx, "if", cdr(exp), 3, 3);

				ifexp = eval(ctx, car(cdr(exp)), env);

				if (!isbool(ifexp))
					err(ctx, ERR_RUNTIME,
						"Boolean value expected in 'if' expression");

				if (ist(ifexp))
					exp = car(cdr(cdr(exp)));
				else
					exp = car(cdr(cdr(cdr(exp))));

				goto eval_start;
			}

			/* (and {bool_exp} {bool_exp} [bool_exp...]) */
			if (car(exp) == ctx->s_and) {
				obj *ret = ctx->tee;

				check_arg_count(ctx, "and", cdr(exp), 2, -1);

				exp = cdr(exp);

				while (!isnil(exp)) {
					obj *b = eval(ctx, car(exp), env);

					if (!isbool(b))
						err(ctx, ERR_RUNTIME,
							"Boolean value expected in 'and' expression");

					if (isf(b)) {
						ret = ctx->eff;
						break;
					}

					exp = cdr(exp);
				}

				return ret;
			}

			/* (or {bool_exp} {bool_exp} [bool_exp...]) */
			if (car(exp) == ctx->s_or) {
				obj *ret = ctx->eff;

				check_arg_count(ctx, "or", cdr(exp), 2, -1);

				exp = cdr(exp);

				while (!isnil(exp)) {
					obj *b = eval(ctx, car(exp), env);

					if (!isbool(b))
						err(ctx, ERR_RUNTIME,
							"Boolean value expected in 'or' expression");

					if (ist(b)) {
						ret = ctx->tee;
						break;
					}

					exp = cdr(exp);
				}

				return ret;
			}

			/* (set! {sym} {exp}) */
			if (car(exp) == ctx->s_setb) {
				obj *sympair, *newval;

				check_arg_count(ctx, "set!", cdr(exp), 2, 2);

				if (!issym(car(cdr(exp))))
					err(ctx, ERR_RUNTIME,
						"Incorrect type for 1st argument of 'set!', symbol expected");

				sympair = assoc(ctx, car(cdr(exp)), env);

				if (isnil(sympair))
					err(ctx, ERR_RUNTIME,
						"Unbound symbol '%s' used as 1st argument of 'set!'",
						symname(car(cdr(exp))));

				newval  = eval(ctx, car(cdr(cdr(exp))), env);
				setcdr(sympair, newval);

				return newval;
			}

			/* (begin [exp...]) */
			if (car(exp) == ctx->s_begin) {
				exp = cdr(exp);

				if (isnil(exp))
					return ctx->nil;

				for (;;) {
                    if (isnil(cdr(exp))) {
						exp = car(exp);
						goto eval_start;
                    }

                    (void)eval(ctx, car(exp), env);

                    exp = cdr(exp);
				}
			}


			/* ({proc} [args]) */
			proc = eval(ctx, car(exp), env);
			vals = eval_list(ctx, cdr(exp), env);

			switch (proc->type) {
				case PRIMITIVE:
					return (*primval(proc))(ctx, vals);

				case LAMBDA:
					if (list_length(vals) != list_length(procargs(proc)))
						err(ctx, ERR_RUNTIME,
							"Expected %d arguments to lambda, %d given",
							list_length(procargs(proc)),
							list_length(vals));

					env = multiple_extend(ctx, procenv(proc), procargs(proc), vals);
					exp = cons(ctx, ctx->s_begin, proccode(proc));
					goto eval_start;

				default:
					err(ctx, ERR_RUNTIME,
						"Invalid function; expected lambda or primitive");
					break;
			}
			break;
	}

	/* Not reached */
	return exp;
}

static obj *eval_list(context *ctx, obj *exps, obj *env)
{
	if (isnil(exps))
		return ctx->nil;

	return cons(ctx, eval(ctx, car(exps), env),
				eval_list(ctx, cdr(exps), env));
}



/**
 *
 * Built-in primitives
 *
 **/

static obj *prim_sum(context *ctx, obj *args)
{
	double sum = 0.0;

	check_arg_count(ctx, "+", args, 2, -1);

	while (!isnil(args)) {
		if (!isnum(car(args)))
			err(ctx, ERR_RUNTIME,
				"Incorrect argument type for '+', number expected");

		sum  += numval(car(args));
		args  = cdr(args);
	}

	return mknum(ctx, sum);
}

static obj *prim_sub(context *ctx, obj *args)
{
	double res = 0.0;

	check_arg_count(ctx, "-", args, 1, -1);

	if (!isnum(car(args)))
		err(ctx, ERR_RUNTIME,
			"Incorrect argument type for '-', number expected");

	res  = numval(car(args));
	args = cdr(args);

	if (isnil(args))
		res = -res;

	while (!isnil(args)) {
		if (!isnum(car(args)))
			err(ctx, ERR_RUNTIME,
				"Incorrect argument type for '-', number expected");

		res  -= numval(car(args));
		args  = cdr(args);
	}

	return mknum(ctx, res);
}

static obj *prim_prod(context *ctx, obj *args)
{
	double prod = 1.0;

	check_arg_count(ctx, "*", args, 2, -1);

	while (!isnil(args)) {
		if (!isnum(car(args)))
			err(ctx, ERR_RUNTIME,
				"Incorrect argument type for '*', number expected");

		prod *= numval(car(args));
		args  = cdr(args);
	}

	return mknum(ctx, prod);
}


#define PRIM_NUMCMP(f, name, op) \
static obj *prim_num ## f(context *ctx, obj *args) \
{ \
	check_arg_count(ctx, name, args, 2, 2); \
	check_arg_types(ctx, name, args, NUM, NUM); \
	\
	return numval(car(args)) op numval(car(cdr(args))) ? ctx->tee : ctx->eff; \
}

PRIM_NUMCMP(eq,  "=",  ==)
PRIM_NUMCMP(leq, "<=", <=)
PRIM_NUMCMP(geq, ">=", >=)
PRIM_NUMCMP(lt,  "<",  <)
PRIM_NUMCMP(gt,  ">",  >)


static obj *prim_list(context *ctx, obj *args)
{
	return args;
}

static obj *prim_cons(context *ctx, obj *args)
{
	check_arg_count(ctx, "cons", args, 2, 2);
	return cons(ctx, car(args), car(cdr(args)));
}

static obj *prim_car(context *ctx, obj *args)
{
	check_arg_count(ctx, "car", args, 1, 1);
	check_arg_types(ctx, "car", args, CONS);

	return car(car(args));
}

static obj *prim_cdr(context *ctx, obj *args)
{
	check_arg_count(ctx, "cdr", args, 1, 1);
	check_arg_types(ctx, "cdr", args, CONS);

	return cdr(car(args));
}

static obj *prim_length(context *ctx, obj *args)
{
	check_arg_count(ctx, "length", args, 1, 1);
	check_arg_types(ctx, "length", args, CONS);

	return mknum(ctx, (double)list_length(car(args)));
}

#define PRIM_LISTREFTAIL(f, name, retval) \
static obj *prim_ ## f(context *ctx, obj *args) \
{ \
	int n = 0, which; \
	\
	check_arg_count(ctx, name, args, 2, 2); \
	check_arg_types(ctx, name, args, CONS, NUM); \
	\
	which = (int)numval(car(cdr(args))); \
	args  = car(args); \
	\
	while (!isnil(args)) { \
		if (n == which) \
			return retval; \
		\
		n++; \
		args = cdr(args); \
	} \
	\
	err(ctx, ERR_RUNTIME, \
		"%s: List index %d out of bounds, maximum is %d", \
		name, which, n - 1); \
	return ctx->nil; \
}

PRIM_LISTREFTAIL(list_ref,  "list-ref",  car(args))
PRIM_LISTREFTAIL(list_tail, "list-tail", args)


#define PRIM_TYPEP(f, name, is) \
static obj *prim_ ## f ## p(context *ctx, obj *args) \
{ \
	check_arg_count(ctx, name, args, 1, 1); \
	return is(car(args)) ? ctx->tee : ctx->eff; \
}

PRIM_TYPEP(null,    "null?",    isnil)
PRIM_TYPEP(number,  "number?",  isnum)
PRIM_TYPEP(symbol,  "symbol?",  issym)
PRIM_TYPEP(pair,    "pair?",    iscons)
PRIM_TYPEP(boolean, "boolean?", isbool)

static obj *prim_procedurep(context *ctx, obj *args)
{
	check_arg_count(ctx, "procedure?", args, 1, 1);
	return (isproc(car(args)) || isprim(car(args))) ? ctx->tee : ctx->eff;
}


#define PRIM_ZEROCMP(f, name, op) \
static obj *prim_ ## f ## p(context *ctx, obj *args) \
{ \
	check_arg_count(ctx, name, args, 1, 1); \
	check_arg_types(ctx, name, args, NUM); \
	\
	return (numval(car(args)) op 0.0) ? ctx->tee : ctx->eff; \
}

PRIM_ZEROCMP(zero,     "zero?",     ==)
PRIM_ZEROCMP(negative, "negative?", <)
PRIM_ZEROCMP(positive, "positive?", >)


#define PRIM_PARITYCMP(f, name, rem) \
static obj *prim_ ## f ## p(context *ctx, obj *args) \
{ \
	check_arg_count(ctx, name, args, 1, 1); \
	check_arg_types(ctx, name, args, NUM); \
	\
	return (fmod(numval(car(args)), 2.0) == rem) ? ctx->tee : ctx->eff; \
}

PRIM_PARITYCMP(odd,  "odd?",  1.0)
PRIM_PARITYCMP(even, "even?", 0.0)


static obj *prim_not(context *ctx, obj *args)
{
	check_arg_count(ctx, "not", args, 1, 1);
	check_arg_types(ctx, "not", args, T);

	return ist(car(args)) ? ctx->eff : ctx->tee;
}

static obj *prim_display(context *ctx, obj *args)
{
	check_arg_count(ctx, "display", args, 1, 1);
	display(ctx, stdout, car(args));

	return ctx->nil;
}

static obj *prim_newline(context *ctx, obj *args)
{
	check_arg_count(ctx, "newline", args, 0, 0);
	fprintf(stdout, "\n");

	return ctx->nil;
}


#define PRIM_MATH(f, name) \
static obj *prim_ ## f (context *ctx, obj *args) \
{ \
	check_arg_count(ctx, name, args, 1, 1); \
	check_arg_types(ctx, name, args, NUM); \
	\
	return mknum(ctx, f(numval(car(args)))); \
}

PRIM_MATH(sin,   "sin")
PRIM_MATH(cos,   "cos")
PRIM_MATH(tan,   "tan")
PRIM_MATH(asin,  "asin")
PRIM_MATH(acos,  "acos")
PRIM_MATH(atan,  "atan")
PRIM_MATH(ceil,  "ceiling")
PRIM_MATH(floor, "floor")
PRIM_MATH(trunc, "truncate")
PRIM_MATH(rint,  "round")
PRIM_MATH(fabs,  "abs")
PRIM_MATH(sqrt,  "sqrt")
PRIM_MATH(exp,   "exp")
PRIM_MATH(log,   "log")

static obj *prim_expt(context *ctx, obj *args)
{
	check_arg_count(ctx, "expt", args, 2, 2);
	check_arg_types(ctx, "expt", args, NUM, NUM);

	return mknum(ctx, pow(numval(car(args)), numval(car(cdr(args)))));
}

#define PRIM_MINMAX(f, name, init, cmp) \
static obj *prim_ ## f(context *ctx, obj *args) \
{ \
	double m = init; \
	\
	check_arg_count(ctx, name, args, 2, -1); \
	\
	while (!isnil(args)) { \
		if (!isnum(car(args))) \
			err(ctx, ERR_RUNTIME, \
				"Incorrect argument type for '%s', number expected", name); \
		\
		if (numval(car(args)) cmp m) \
			m = numval(car(args)); \
		\
		args = cdr(args); \
	} \
	\
	return mknum(ctx, m); \
}

PRIM_MINMAX(min, "min", DBL_MAX, <)
PRIM_MINMAX(max, "max", DBL_MIN, >)


static obj *prim_gcd(context *ctx, obj *args)
{
	double a, b;

	check_arg_count(ctx, "gcd", args, 2, 2);
	check_arg_types(ctx, "gcd", args, NUM, NUM);

	a = numval(car(args));
	b = numval(car(cdr(args)));

	if (fabs(a) < DBL_EPSILON)
		return mknum(ctx, b);

	while (fabs(b) > DBL_EPSILON) {
		if (a > b)
			a -= b;
		else
			b -= a;
	}

	return mknum(ctx, a);
}




static const char scheme_library_code[] = "			\
(define map (lambda (func list)						\
	(if (null? list)								\
		'()											\
		(cons (func (car list))						\
			(map func (cdr list))))))				\
													\
(define reduce (lambda (func list id)				\
	(if (null? list)								\
		id											\
		(func (car list)							\
			(reduce func (cdr list) id)))))			\
													\
(define e  2.71828182845904523536028747135266250)	\
(define pi 3.14159265358979323846264338327950288)	\
";

void register_primitive(context *ctx, const char *name, primitive prim)
{
	(void)extend_top(ctx, intern(ctx, name), mkprimitive(ctx, prim));
}



/**
 *
 * Context handling
 *
 **/

context *create_context(void)
{
	/* Allocate memory for new context */
	context *ctx = (context *)GC_MALLOC(sizeof(context));

	if (!ctx)
		return NULL;

	(void)memset(ctx, 0, sizeof(context));

	/* Set up basic symbols & environment */
	ctx->nil      = mknil(ctx);
	ctx->symbols  = ctx->nil;
	ctx->top_env  = cons(ctx, cons(ctx, ctx->nil, ctx->nil), ctx->nil);
	ctx->tee      = mkt(ctx);
	ctx->eff      = mkf(ctx);
	ctx->s_quote  = intern(ctx, "quote");
	ctx->s_define = intern(ctx, "define");
	ctx->s_lambda = intern(ctx, "lambda");
	ctx->s_setb   = intern(ctx, "set!");
	ctx->s_begin  = intern(ctx, "begin");
	ctx->s_if     = intern(ctx, "if");
	ctx->s_and    = intern(ctx, "and");
	ctx->s_or     = intern(ctx, "or");

	/* Set up built-in primitives */
	register_primitive(ctx, "+", prim_sum);
	register_primitive(ctx, "-", prim_sub);
	register_primitive(ctx, "*", prim_prod);
	register_primitive(ctx, "=", prim_numeq);
	register_primitive(ctx, "<=", prim_numleq);
	register_primitive(ctx, ">=", prim_numgeq);
	register_primitive(ctx, "<", prim_numlt);
	register_primitive(ctx, ">", prim_numgt);
	register_primitive(ctx, "list", prim_list);
	register_primitive(ctx, "cons", prim_cons);
	register_primitive(ctx, "car", prim_car);
	register_primitive(ctx, "cdr", prim_cdr);
	register_primitive(ctx, "length", prim_length);
	register_primitive(ctx, "list-ref", prim_list_ref);
	register_primitive(ctx, "list-tail", prim_list_tail);

	register_primitive(ctx, "null?", prim_nullp);
	register_primitive(ctx, "number?", prim_numberp);
	register_primitive(ctx, "symbol?", prim_symbolp);
	register_primitive(ctx, "pair?", prim_pairp);
	register_primitive(ctx, "procedure?", prim_procedurep);
	register_primitive(ctx, "boolean?", prim_booleanp);
	register_primitive(ctx, "zero?", prim_zerop);
	register_primitive(ctx, "negative?", prim_negativep);
	register_primitive(ctx, "positive?", prim_positivep);
	register_primitive(ctx, "odd?", prim_oddp);
	register_primitive(ctx, "even?", prim_evenp);

	register_primitive(ctx, "not", prim_not);
	register_primitive(ctx, "display", prim_display);
	register_primitive(ctx, "newline", prim_newline);

	register_primitive(ctx, "sin", prim_sin);
	register_primitive(ctx, "cos", prim_cos);
	register_primitive(ctx, "tan", prim_tan);
	register_primitive(ctx, "asin", prim_asin);
	register_primitive(ctx, "acos", prim_acos);
	register_primitive(ctx, "atan", prim_atan);
	register_primitive(ctx, "ceiling", prim_ceil);
	register_primitive(ctx, "floor", prim_floor);
	register_primitive(ctx, "truncate", prim_trunc);
	register_primitive(ctx, "round", prim_rint);
	register_primitive(ctx, "abs", prim_fabs);
	register_primitive(ctx, "sqrt", prim_sqrt);
	register_primitive(ctx, "exp", prim_exp);
	register_primitive(ctx, "log", prim_log);
	register_primitive(ctx, "expt", prim_expt);
	register_primitive(ctx, "min", prim_min);
	register_primitive(ctx, "max", prim_max);
	register_primitive(ctx, "gcd", prim_gcd);

	/* Set up built-in library */
	(void)eval(ctx, parse(ctx, scheme_library_code), ctx->top_env);

	return ctx;
}


void destroy_context(context *ctx)
{
	if (!ctx)
		return;

	if (ctx->p)
		GC_FREE(ctx->p);

	GC_FREE(ctx);
	ctx = NULL;
}


obj *eval_file(context *ctx, const char *fname)
{
	struct stat  sb;
	char        *source = NULL;
	int          fd     = 0;
	obj         *ret    = ctx->nil;

	sb.st_size = 0; /* Silly compiler */

	if (setjmp(ctx->err_handler) != 0) {
		/* An error has occurred */

		if (ctx->p) {
			GC_FREE(ctx->p);
			ctx->p = NULL;
		}

		if (source)
			(void)munmap(source, (size_t)sb.st_size);

		if (fd)
			(void)close(fd);

		return ctx->nil;
	}

	if ((fd = open(fname, O_RDONLY)) == -1)
		err(ctx, ERR_FATAL, "Unable to open file '%s'", fname);

	if (stat(fname, &sb) == -1)
		err(ctx, ERR_FATAL, "stat() failed");

	source = (char *)mmap(NULL, (size_t)sb.st_size, PROT_READ, MAP_SHARED, fd, 0);

	if (source == MAP_FAILED)
		err(ctx, ERR_FATAL, "mmap() failed");

	ret = eval(ctx, parse(ctx, source), ctx->top_env);

	(void)munmap(source, (size_t)sb.st_size);
	(void)close(fd);

	return ret;
}


obj *eval_string(context *ctx, const char *source)
{
	if (setjmp(ctx->err_handler) != 0) {
		if (ctx->p) {
			GC_FREE(ctx->p);
			ctx->p = NULL;
		}

		return ctx->nil;
	}

	return eval(ctx, parse(ctx, source), ctx->top_env);
}


void scheme_init(void)
{
	GC_INIT();
}

