#ifndef __SCHEME_H__
#define __SCHEME_H__

#include <setjmp.h>
#include <sys/types.h>

#ifdef __cplusplus
extern "C" {
#endif

struct __context;

enum obj_type { NIL, T, F, NUM, SYM, STR, CONS, LAMBDA, PRIMITIVE };

#define OBJ_HEAD \
	enum obj_type type; \
	size_t        at_line, at_column

typedef struct { OBJ_HEAD; } obj;
typedef struct { OBJ_HEAD; obj *car, *cdr; } obj_cons;
typedef struct { OBJ_HEAD; double val; } obj_num;
typedef struct { OBJ_HEAD; char *sym; } obj_sym;
typedef struct { OBJ_HEAD; char *str; } obj_str;
typedef struct { OBJ_HEAD; obj *(*prim)(struct __context *, obj *); } obj_prim;
typedef struct { OBJ_HEAD; obj *args, *code, *env; } obj_proc;

typedef obj *(*primitive)(struct __context *, obj *);

#define car(O)          (((obj_cons *)(O))->car)
#define cdr(O)          (((obj_cons *)(O))->cdr)
#define setcar(O, X)    ((((obj_cons *)(O))->car) = (X))
#define setcdr(O, X)    ((((obj_cons *)(O))->cdr) = (X))
#define numval(O)       (((obj_num  *)(O))->val)
#define symname(O)      (((obj_sym  *)(O))->sym)
#define strval(O)       (((obj_str  *)(O))->str)
#define primval(O)      (((obj_prim *)(O))->prim)
#define procargs(O)     (((obj_proc *)(O))->args)
#define proccode(O)     (((obj_proc *)(O))->code)
#define procenv(O)      (((obj_proc *)(O))->env)

#define isnil(O)        ((O)->type == NIL)
#define ist(O)          ((O)->type == T)
#define isf(O)          ((O)->type == F)
#define isbool(O)       (ist(O) || isf(O))
#define isnum(O)        ((O)->type == NUM)
#define issym(O)        ((O)->type == SYM)
#define isstr(O)        ((O)->type == STR)
#define iscons(O)       ((O)->type == CONS)
#define isproc(O)       ((O)->type == LAMBDA)
#define isprim(O)       ((O)->type == PRIMITIVE)

enum tok_type {
	TOK_EOI, TOK_LPAR, TOK_RPAR, TOK_QUOTE, TOK_DOT,
	TOK_NUM, TOK_SYM, TOK_STR, TOK_T, TOK_F
};

typedef struct {
	const char    *source, *ptr, *ptr_prev, *line_start, *line_start_prev;
	size_t         line, line_prev, column, column_prev;
	size_t         length;
	
	enum tok_type  tok;
	double         tok_num;
	char           tok_sym[40];
	char           *tok_str;
	char           error[40];
} parser;

enum err_type { ERR_FATAL = 1, ERR_SYNTAX, ERR_RUNTIME };

typedef struct __context {	
	/* Context symbols */
	obj     *symbols, *top_env, *nil, *tee, *eff;
	obj     *s_quote, *s_lambda, *s_define, *s_setb, *s_begin;
	obj     *s_if, *s_and, *s_or;
	
	/* Current parser context, if any */
	parser  *p;
	
	/* Context error handler */
	jmp_buf  err_handler;
	size_t   err_line, err_column;	/* Location data for runtime errors */
	
	/* Context error string */
	char error[512];
} context;

extern void     err(context *ctx, enum err_type type, const char *fmt, ...);
extern obj     *cons(context *ctx, obj *_car, obj *_cdr);
extern obj     *mknil(context *ctx);
extern obj     *mkt(context *ctx);
extern obj     *mkf(context *ctx);
extern obj     *mknum(context *ctx, double n);
extern obj     *mksym(context *ctx, const char *sym);
extern obj     *mkstr(context *ctx, const char *str);
extern obj     *mkprimitive(context *ctx, primitive prim);
extern obj     *mkproc(context *ctx, obj *args, obj *code, obj *env);
extern obj     *assoc(context *ctx, obj *key, obj *alist);
extern obj     *parse(context *ctx, const char *src);
extern void     display(context *ctx, FILE *ofp, obj *op);
extern obj     *eval(context *ctx, obj *exp, obj *env);
extern void     check_arg_count(context *ctx, const char *fn, obj *args, int min, int max);
extern void     check_arg_types(context *ctx, const char *fn, obj *args, ...);
extern int      list_length(obj *list);
extern void     register_primitive(context *ctx, const char *name, primitive prim);
extern context *create_context(void);
extern void     destroy_context(context *ctx);
extern obj     *eval_file(context *ctx, const char *fname);
extern obj     *eval_string(context *ctx, const char *source);
extern void     scheme_init(void);

#ifdef __cplusplus
}	/* extern "C" */
#endif		

#endif /* __SCHEME_H__ */
